# Copyright Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from os import environ
from typing import Optional

from opentelemetry.sdk.resources import Resource

from splunk_otel.util import _is_truthy

logger = logging.getLogger(__name__)
_DEFAULT_CALL_STACK_INTERVAL_MILLIS = 1_000


def _sanitize_interval(interval):
    if isinstance(interval, int):
        if interval < 1:
            logger.warning(
                "call stack interval has to be positive, got %s, defaulting to %s",
                interval,
                _DEFAULT_CALL_STACK_INTERVAL_MILLIS,
            )
            return _DEFAULT_CALL_STACK_INTERVAL_MILLIS

        return interval

    logger.warning(
        "call stack interval not an integer, defaulting to %s",
        _DEFAULT_CALL_STACK_INTERVAL_MILLIS,
    )
    return _DEFAULT_CALL_STACK_INTERVAL_MILLIS


class _Options:
    resource: Resource
    endpoint: str
    call_stack_interval_millis: int

    def __init__(
        self,
        resource: Resource,
        endpoint: Optional[str] = None,
        call_stack_interval_millis: Optional[int] = None,
        include_internal_stacks: Optional[bool] = None,
    ):
        self.resource = resource
        self.endpoint = _Options._get_endpoint(endpoint)
        self.call_stack_interval_millis = _Options._get_call_stack_interval(
            call_stack_interval_millis
        )
        self.include_internal_stacks = _Options._include_internal_stacks(
            include_internal_stacks
        )

    @staticmethod
    def _get_endpoint(endpoint: Optional[str]) -> str:
        if not endpoint:
            endpoint = environ.get("SPLUNK_PROFILER_LOGS_ENDPOINT")

        if not endpoint:
            endpoint = environ.get("OTEL_EXPORTER_OTLP_ENDPOINT")

        return endpoint or "http://localhost:4317"

    @staticmethod
    def _get_call_stack_interval(interval: Optional[int]) -> int:
        if not interval:
            interval = environ.get("SPLUNK_PROFILER_CALL_STACK_INTERVAL")

            if interval:
                return _sanitize_interval(int(interval))

            return _DEFAULT_CALL_STACK_INTERVAL_MILLIS

        return _sanitize_interval(interval)

    @staticmethod
    def _include_internal_stacks(include: Optional[bool]) -> bool:
        if include is None:
            return _is_truthy(environ.get("SPLUNK_PROFILER_INCLUDE_INTERNAL_STACKS"))

        return include



# Copyright Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from functools import partial
from os import environ
from typing import Callable, Collection, Dict, List, Optional, Tuple, Union

from opentelemetry.environment_variables import OTEL_TRACES_EXPORTER
from opentelemetry.instrumentation.propagators import ResponsePropagator
from opentelemetry.instrumentation.version import (
    __version__ as auto_instrumentation_version,
)
from opentelemetry.sdk.environment_variables import (
    OTEL_ATTRIBUTE_COUNT_LIMIT,
    OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT,
    OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT,
    OTEL_LINK_ATTRIBUTE_COUNT_LIMIT,
    OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT,
    OTEL_SPAN_EVENT_COUNT_LIMIT,
    OTEL_SPAN_LINK_COUNT_LIMIT,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace.export import SpanExporter
from pkg_resources import iter_entry_points

from splunk_otel.environment_variables import (
    _SPLUNK_ACCESS_TOKEN,
    _SPLUNK_TRACE_RESPONSE_HEADER_ENABLED,
)
from splunk_otel.propagators import _ServerTimingResponsePropagator
from splunk_otel.symbols import (
    _DEFAULT_EXPORTERS,
    _DEFAULT_MAX_ATTR_LENGTH,
    _DEFAULT_OTEL_SERVICE_NAME,
    _DEFAULT_SERVICE_NAME,
    _DEFAULT_SPAN_LINK_COUNT_LIMIT,
    _EXPORTER_OTLP,
    _EXPORTER_OTLP_GRPC,
    _KNOWN_EXPORTER_PACKAGES,
    _LIMIT_UNSET_VALUE,
    _NO_SERVICE_NAME_WARNING,
    _SERVICE_NAME_ATTR,
    _SPLUNK_DISTRO_VERSION_ATTR,
    _TELEMETRY_VERSION_ATTR,
)
from splunk_otel.util import _is_truthy_str
from splunk_otel.version import __version__

_SpanExporterFactory = Callable[["_Options"], SpanExporter]
_SpanExporterClass = Callable[..., SpanExporter]

logger = logging.getLogger("options")


class _Options:
    span_exporter_factories: Collection[_SpanExporterFactory]
    access_token: Optional[str]
    resource: Resource
    response_propagation: bool
    response_propagator: Optional[ResponsePropagator]

    def __init__(
        self,
        service_name: Optional[str] = None,
        span_exporter_factories: Optional[Collection[_SpanExporterFactory]] = None,
        access_token: Optional[str] = None,
        resource_attributes: Optional[Dict[str, Union[str, bool, int, float]]] = None,
        trace_response_header_enabled: Optional[bool] = None,
    ):
        # todo: remove this side effect
        _set_default_env()

        self.access_token = _resolve_access_token(access_token)
        self.response_propagator = _get_response_propagator(trace_response_header_enabled)
        self.resource = _create_resource(service_name, resource_attributes)
        self.span_exporter_factories = _get_span_exporter_factories(
            span_exporter_factories
        )


def _resolve_access_token(access_token: Optional[str]) -> Optional[str]:
    if not access_token:
        access_token = environ.get(_SPLUNK_ACCESS_TOKEN)
    return access_token or None


def _get_response_propagator(
    enabled: Optional[bool],
) -> Optional[ResponsePropagator]:
    if enabled is None:
        enabled = _is_truthy_str(
            environ.get(_SPLUNK_TRACE_RESPONSE_HEADER_ENABLED, "true")
        )
    if enabled:
        return _ServerTimingResponsePropagator()
    return None


def _create_resource(
    service_name: Optional[str],
    attributes: Optional[Dict[str, Union[str, bool, int, float]]],
) -> Resource:
    attributes = attributes or {}
    if service_name:
        attributes[_SERVICE_NAME_ATTR] = service_name
    attributes.update(
        {
            _TELEMETRY_VERSION_ATTR: auto_instrumentation_version,
            _SPLUNK_DISTRO_VERSION_ATTR: __version__,
        }
    )
    resource = Resource.create(attributes)
    if (
        resource.attributes.get(_SERVICE_NAME_ATTR, _DEFAULT_OTEL_SERVICE_NAME)
        == _DEFAULT_OTEL_SERVICE_NAME
    ):
        logger.warning(_NO_SERVICE_NAME_WARNING)
        resource = resource.merge(Resource({_SERVICE_NAME_ATTR: _DEFAULT_SERVICE_NAME}))
    return resource


def _get_span_exporter_factories(
    factories: Optional[Collection[_SpanExporterFactory]],
) -> Collection[_SpanExporterFactory]:
    if factories:
        return factories

    exporter_names = _get_span_exporter_names_from_env()
    return _import_span_exporter_factories(exporter_names)


def _set_default_env() -> None:
    defaults = {
        OTEL_ATTRIBUTE_COUNT_LIMIT: _LIMIT_UNSET_VALUE,
        OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT: _LIMIT_UNSET_VALUE,
        OTEL_SPAN_EVENT_COUNT_LIMIT: _LIMIT_UNSET_VALUE,
        OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT: _LIMIT_UNSET_VALUE,
        OTEL_LINK_ATTRIBUTE_COUNT_LIMIT: _LIMIT_UNSET_VALUE,
        OTEL_SPAN_LINK_COUNT_LIMIT: str(_DEFAULT_SPAN_LINK_COUNT_LIMIT),
        OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT: str(_DEFAULT_MAX_ATTR_LENGTH),
    }

    for key, value in defaults.items():
        if key not in environ:
            environ[key] = value


def _get_span_exporter_names_from_env() -> Collection[Tuple[str, str]]:
    exporters_env = environ.get(OTEL_TRACES_EXPORTER, "").strip() or _DEFAULT_EXPORTERS

    exporters: List[Tuple[str, str]] = []
    if not _is_truthy_str(exporters_env):
        return exporters

    # exporters are known by different names internally by Python Otel SDK.
    # Here we create a mapping of user provided names to internal names so
    # that we can provide helpful error messages later.
    for name in exporters_env.split(","):
        if name == _EXPORTER_OTLP:
            exporters.append((_EXPORTER_OTLP, _EXPORTER_OTLP_GRPC))
        else:
            exporters.append(
                (
                    name,
                    name,
                )
            )
    return exporters


def _import_span_exporter_factories(
    exporter_names: Collection[Tuple[str, str]],
) -> Collection[_SpanExporterFactory]:
    factories = []
    entry_points = {
        ep.name: ep for ep in iter_entry_points("opentelemetry_traces_exporter")
    }

    for name, internal_name in exporter_names:
        if internal_name not in entry_points:
            package = _KNOWN_EXPORTER_PACKAGES.get(internal_name)
            if package:
                help_msg = f"please make sure {package} is installed"
            else:
                help_msg = "please make sure the relevant exporter package is installed."
            raise ValueError(f'exporter "{name} ({internal_name})" not found. {help_msg}')

        exporter_class: _SpanExporterClass = entry_points[internal_name].load()
        if internal_name == _EXPORTER_OTLP_GRPC:
            factory = _otlp_factory
        else:
            factory = _generic_exporter
        factories.append(partial(factory, exporter_class))
    return factories


def _generic_exporter(
    exporter: _SpanExporterClass,
    options: _Options,  # pylint: disable=unused-argument
) -> SpanExporter:
    return exporter()


def _otlp_factory(exporter: _SpanExporterClass, options: "_Options") -> SpanExporter:
    return exporter(headers=(("x-sf-token", options.access_token),))



# Copyright Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Collection, Dict, Optional, Union

from opentelemetry import trace
from opentelemetry.instrumentation.environment_variables import (
    OTEL_PYTHON_DISABLED_INSTRUMENTATIONS,
)
from opentelemetry.instrumentation.propagators import set_global_response_propagator
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from pkg_resources import iter_entry_points

from splunk_otel.options import _Options, _SpanExporterFactory
from splunk_otel.util import _is_truthy

logger = logging.getLogger(__name__)


def start_tracing(
    service_name: Optional[str] = None,
    span_exporter_factories: Optional[Collection[_SpanExporterFactory]] = None,
    access_token: Optional[str] = None,
    resource_attributes: Optional[Dict[str, Union[str, bool, int, float]]] = None,
    trace_response_header_enabled: Optional[bool] = None,
) -> trace.TracerProvider:
    enabled = os.environ.get("OTEL_TRACE_ENABLED", True)
    if not _is_truthy(enabled):
        logger.info("tracing has been disabled with OTEL_TRACE_ENABLED=%s", enabled)
        return None

    options = _Options(
        service_name,
        span_exporter_factories,
        access_token,
        resource_attributes,
        trace_response_header_enabled,
    )
    try:
        provider = _configure_tracing(options)
        _load_instrumentors()
        return provider
    except Exception as error:  # pylint:disable=broad-except
        logger.error("tracing could not be enabled: %s", error)
        return trace.NoOpTracerProvider()


def _configure_tracing(options: _Options) -> TracerProvider:
    provider = TracerProvider(resource=options.resource)
    set_global_response_propagator(options.response_propagator)  # type: ignore
    trace.set_tracer_provider(provider)
    for factory in options.span_exporter_factories:
        provider.add_span_processor(BatchSpanProcessor(factory(options)))
    return provider


def _load_instrumentors() -> None:
    package_to_exclude = os.environ.get(OTEL_PYTHON_DISABLED_INSTRUMENTATIONS, "").split(
        ","
    )
    package_to_exclude = [p.strip() for p in package_to_exclude]

    for entry_point in iter_entry_points("opentelemetry_instrumentor"):
        try:
            if entry_point.name in package_to_exclude:
                logger.debug("Instrumentation skipped for library %s", entry_point.name)
                continue
            entry_point.load()().instrument()
            logger.debug("Instrumented %s", entry_point.name)
        except Exception as exc:  # pylint: disable=broad-except
            logger.exception("Instrumenting of %s failed", entry_point.name)
            raise exc



# pylint: skip-file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: profile.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rprofile.proto\x12\x12perftools.profiles"\xd5\x03\n\x07Profile\x12\x32\n\x0bsample_type\x18\x01 \x03(\x0b\x32\x1d.perftools.profiles.ValueType\x12*\n\x06sample\x18\x02 \x03(\x0b\x32\x1a.perftools.profiles.Sample\x12,\n\x07mapping\x18\x03 \x03(\x0b\x32\x1b.perftools.profiles.Mapping\x12.\n\x08location\x18\x04 \x03(\x0b\x32\x1c.perftools.profiles.Location\x12.\n\x08\x66unction\x18\x05 \x03(\x0b\x32\x1c.perftools.profiles.Function\x12\x14\n\x0cstring_table\x18\x06 \x03(\t\x12\x13\n\x0b\x64rop_frames\x18\x07 \x01(\x03\x12\x13\n\x0bkeep_frames\x18\x08 \x01(\x03\x12\x12\n\ntime_nanos\x18\t \x01(\x03\x12\x16\n\x0e\x64uration_nanos\x18\n \x01(\x03\x12\x32\n\x0bperiod_type\x18\x0b \x01(\x0b\x32\x1d.perftools.profiles.ValueType\x12\x0e\n\x06period\x18\x0c \x01(\x03\x12\x0f\n\x07\x63omment\x18\r \x03(\x03\x12\x1b\n\x13\x64\x65\x66\x61ult_sample_type\x18\x0e \x01(\x03"\'\n\tValueType\x12\x0c\n\x04type\x18\x01 \x01(\x03\x12\x0c\n\x04unit\x18\x02 \x01(\x03"V\n\x06Sample\x12\x13\n\x0blocation_id\x18\x01 \x03(\x04\x12\r\n\x05value\x18\x02 \x03(\x03\x12(\n\x05label\x18\x03 \x03(\x0b\x32\x19.perftools.profiles.Label"@\n\x05Label\x12\x0b\n\x03key\x18\x01 \x01(\x03\x12\x0b\n\x03str\x18\x02 \x01(\x03\x12\x0b\n\x03num\x18\x03 \x01(\x03\x12\x10\n\x08num_unit\x18\x04 \x01(\x03"\xdd\x01\n\x07Mapping\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x14\n\x0cmemory_start\x18\x02 \x01(\x04\x12\x14\n\x0cmemory_limit\x18\x03 \x01(\x04\x12\x13\n\x0b\x66ile_offset\x18\x04 \x01(\x04\x12\x10\n\x08\x66ilename\x18\x05 \x01(\x03\x12\x10\n\x08\x62uild_id\x18\x06 \x01(\x03\x12\x15\n\rhas_functions\x18\x07 \x01(\x08\x12\x15\n\rhas_filenames\x18\x08 \x01(\x08\x12\x18\n\x10has_line_numbers\x18\t \x01(\x08\x12\x19\n\x11has_inline_frames\x18\n \x01(\x08"v\n\x08Location\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x12\n\nmapping_id\x18\x02 \x01(\x04\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x01(\x04\x12&\n\x04line\x18\x04 \x03(\x0b\x32\x18.perftools.profiles.Line\x12\x11\n\tis_folded\x18\x05 \x01(\x08")\n\x04Line\x12\x13\n\x0b\x66unction_id\x18\x01 \x01(\x04\x12\x0c\n\x04line\x18\x02 \x01(\x03"_\n\x08\x46unction\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x01(\x03\x12\x13\n\x0bsystem_name\x18\x03 \x01(\x03\x12\x10\n\x08\x66ilename\x18\x04 \x01(\x03\x12\x12\n\nstart_line\x18\x05 \x01(\x03\x42-\n\x1d\x63om.google.perftools.profilesB\x0cProfileProtob\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "profile_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS is False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"\n\035com.google.perftools.profilesB\014ProfileProto"
    )
    _globals["_PROFILE"]._serialized_start = 38
    _globals["_PROFILE"]._serialized_end = 507
    _globals["_VALUETYPE"]._serialized_start = 509
    _globals["_VALUETYPE"]._serialized_end = 548
    _globals["_SAMPLE"]._serialized_start = 550
    _globals["_SAMPLE"]._serialized_end = 636
    _globals["_LABEL"]._serialized_start = 638
    _globals["_LABEL"]._serialized_end = 702
    _globals["_MAPPING"]._serialized_start = 705
    _globals["_MAPPING"]._serialized_end = 926
    _globals["_LOCATION"]._serialized_start = 928
    _globals["_LOCATION"]._serialized_end = 1046
    _globals["_LINE"]._serialized_start = 1048
    _globals["_LINE"]._serialized_end = 1089
    _globals["_FUNCTION"]._serialized_start = 1091
    _globals["_FUNCTION"]._serialized_end = 1186
# @@protoc_insertion_point(module_scope)



# Copyright Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import base64
import gzip
import logging
import os
import sys
import threading
import time
import traceback
from collections import OrderedDict
from traceback import StackSummary
from typing import Dict, Optional, Union

import opentelemetry.context
import wrapt
from opentelemetry._logs import SeverityNumber
from opentelemetry.context import Context
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.sdk._logs import LoggerProvider, LogRecord
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.trace import TraceFlags
from opentelemetry.trace.propagation import _SPAN_KEY

from splunk_otel.options import _create_resource
from splunk_otel.profiling import profile_pb2
from splunk_otel.profiling.options import _Options

logger = logging.getLogger(__name__)


class Profiler:
    def __init__(self):
        self.condition = threading.Condition(threading.Lock())
        self.thread_states = {}

        self.resource = None
        self.call_stack_interval_millis = None
        self.include_internal_stacks = None
        self.running = False

        self.logger_provider = None
        self.exporter = None
        self.batch_processor = None

    def setup(
        self, resource, endpoint, call_stack_interval_millis, include_internal_stacks
    ):
        # pylint: disable=import-outside-toplevel
        from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter

        self.resource = resource
        self.call_stack_interval_millis = call_stack_interval_millis
        self.include_internal_stacks = include_internal_stacks
        self.running = True

        self.exporter = OTLPLogExporter(endpoint)
        self.batch_processor = BatchLogRecordProcessor(self.exporter)
        self.logger_provider = LoggerProvider(resource=resource)
        self.logger_provider.add_log_record_processor(self.batch_processor)

    def loop(self):
        ignored_thread_ids = self.get_ignored_thread_ids()

        profiling_logger = self.logger_provider.get_logger("otel.profiling", "0.1.0")

        call_stack_interval_seconds = self.call_stack_interval_millis / 1e3
        # In case the processing takes more than the given interval, default to the smallest allowed interval
        min_call_stack_interval_seconds = 1 / 1e3
        while self.is_running():
            time_begin = time.time()
            timestamp = int(time_begin * 1e9)

            stacktraces = _collect_stacktraces(ignored_thread_ids)

            log_record = self.to_log_record(
                stacktraces,
                timestamp,
            )

            profiling_logger.emit(log_record)

            processing_time = time.time() - time_begin
            wait_for = max(
                call_stack_interval_seconds - processing_time,
                min_call_stack_interval_seconds,
            )

            with self.condition:
                self.condition.wait(wait_for)

        # Hack around batch log processor getting stuck on application exits when the profiling endpoint is not reachable
        # Could be related: https://github.com/open-telemetry/opentelemetry-python/issues/2284
        # pylint: disable-next=protected-access
        self.exporter._shutdown = True
        self.logger_provider.shutdown()
        with self.condition:
            self.condition.notify_all()

    def get_ignored_thread_ids(self):
        if self.include_internal_stacks:
            return []

        ignored_ids = [threading.get_ident()]

        log_processor_thread_id = self.extract_log_processor_thread_id()

        if log_processor_thread_id:
            ignored_ids.append(log_processor_thread_id)

        return ignored_ids

    def extract_log_processor_thread_id(self):
        if hasattr(self.batch_processor, "_worker_thread"):
            # pylint: disable-next=protected-access
            worker_thread = self.batch_processor._worker_thread

            if isinstance(worker_thread, threading.Thread):
                return worker_thread.ident

        return None

    def to_log_record(self, stacktraces, timestamp_unix_nanos):
        encoded = self.get_cpu_profile(stacktraces, timestamp_unix_nanos)
        encoded_profile = base64.b64encode(encoded).decode()

        frame_count = 0
        for stacktrace in stacktraces:
            frame_count += len(stacktrace["frames"])

        return LogRecord(
            timestamp=timestamp_unix_nanos,
            trace_id=0,
            span_id=0,
            trace_flags=TraceFlags(0x01),
            severity_number=SeverityNumber.UNSPECIFIED,
            body=encoded_profile,
            resource=self.resource,
            attributes={
                "profiling.data.format": "pprof-gzip-base64",
                "profiling.data.type": "cpu",
                "com.splunk.sourcetype": "otel.profiling",
                "profiling.data.total.frame.count": frame_count,
            },
        )

    def get_cpu_profile(self, stacktraces, timestamp_unix_nanos):
        str_table = StringTable()
        locations_table = OrderedDict()
        functions_table = OrderedDict()

        def get_function(file_name, function_name):
            key = f"{file_name}:{function_name}"
            fun = functions_table.get(key)

            if fun is None:
                name_id = str_table.index(function_name)
                fun = profile_pb2.Function()
                fun.id = len(functions_table) + 1
                fun.name = name_id
                fun.system_name = name_id
                fun.filename = str_table.index(file_name)
                functions_table[key] = fun

            return fun

        def get_line(file_name, function_name, line_no):
            line = profile_pb2.Line()
            line.function_id = get_function(file_name, function_name).id
            line.line = line_no if line_no != 0 else -1
            return line

        def get_location(frame):
            (file_name, function_name, line_no) = frame
            key = f"{file_name}:{function_name}:{line_no}"
            location = locations_table.get(key)

            if location is None:
                location = profile_pb2.Location()
                location.id = len(locations_table) + 1
                location.line.append(get_line(file_name, function_name, line_no))
                locations_table[key] = location

            return location

        timestamp_unix_millis = int(timestamp_unix_nanos / 1e6)

        timestamp_key = str_table.index("source.event.time")
        trace_id_key = str_table.index("trace_id")
        span_id_key = str_table.index("span_id")
        thread_id_key = str_table.index("thread.id")
        event_period_key = str_table.index("source.event.period")

        pb_profile = profile_pb2.Profile()

        event_period_label = profile_pb2.Label()
        event_period_label.key = event_period_key
        event_period_label.num = self.call_stack_interval_millis

        samples = []
        for stacktrace in stacktraces:
            thread_id = stacktrace["tid"]

            timestamp_label = profile_pb2.Label()
            timestamp_label.key = timestamp_key
            timestamp_label.num = timestamp_unix_millis

            thread_id_label = profile_pb2.Label()
            thread_id_label.key = thread_id_key
            thread_id_label.num = thread_id

            labels = [timestamp_label, event_period_label, thread_id_label]

            trace_context = self.thread_states.get(thread_id)
            if trace_context:
                (trace_id, span_id) = trace_context

                trace_id_label = profile_pb2.Label()
                trace_id_label.key = trace_id_key
                trace_id_label.str = str_table.index(f"{trace_id:016x}")
                labels.append(trace_id_label)

                span_id_label = profile_pb2.Label()
                span_id_label.key = span_id_key
                span_id_label.str = str_table.index(f"{span_id:08x}")
                labels.append(span_id_label)

            sample = profile_pb2.Sample()

            location_ids = []

            for frame in reversed(stacktrace["frames"]):
                location_ids.append(get_location(frame).id)

            sample.location_id.extend(location_ids)
            sample.label.extend(labels)

            samples.append(sample)

        pb_profile.sample.extend(samples)
        pb_profile.string_table.extend(str_table.keys())
        pb_profile.function.extend(list(functions_table.values()))
        pb_profile.location.extend(list(locations_table.values()))

        return gzip.compress(pb_profile.SerializeToString())

    def is_running(self):
        return self.running

    def force_flush(self):
        if self.logger_provider:
            self.logger_provider.force_flush()

    def make_wrapped_context_attach(self):
        def _wrapped_context_attach(wrapped, _instance, args, kwargs):
            token = wrapped(*args, **kwargs)

            maybe_context = args[0] if args else None

            if maybe_context:
                span = maybe_context.get(_SPAN_KEY)

                if span:
                    thread_id = threading.get_ident()
                    context = span.get_span_context()
                    self.thread_states[thread_id] = (
                        context.trace_id,
                        context.span_id,
                    )

            return token

        return _wrapped_context_attach

    def make_wrapped_context_detach(self):
        def _wrapped_context_detach(wrapped, _instance, args, kwargs):
            token = args[0] if args else None

            if token:
                prev = token.old_value
                thread_id = threading.get_ident()
                if isinstance(prev, Context):
                    span = prev.get(_SPAN_KEY)

                    if span:
                        context = span.get_span_context()
                        self.thread_states[thread_id] = (
                            context.trace_id,
                            context.span_id,
                        )
                    else:
                        self.thread_states[thread_id] = None
                else:
                    self.thread_states[thread_id] = None
            return wrapped(*args, **kwargs)

        return _wrapped_context_detach


_profiler = Profiler()


def get_profiler():
    return _profiler


class StringTable:
    def __init__(self):
        self.strings = OrderedDict()

    def index(self, token):
        idx = self.strings.get(token)

        if idx:
            return idx

        idx = len(self.strings)
        self.strings[token] = idx
        return idx

    def keys(self):
        return list(self.strings.keys())


def extract_stack(frame):
    stack = StackSummary.extract(
        traceback.walk_stack(frame), limit=None, lookup_lines=False
    )
    stack.reverse()
    return stack


def _collect_stacktraces(ignored_thread_ids):
    stacktraces = []

    frames = sys._current_frames()

    for thread_id, frame in frames.items():
        if thread_id in ignored_thread_ids:
            continue

        stacktrace_frames = []
        stack = extract_stack(frame)
        for sf in stack:
            stacktrace_frames.append((sf.filename, sf.name, sf.lineno))
        stacktrace = {
            "frames": stacktrace_frames,
            "tid": thread_id,
        }
        stacktraces.append(stacktrace)
    return stacktraces


def _start_profiler_thread(profiler):
    profiler_thread = threading.Thread(
        name="splunk-otel-profiler", target=profiler.loop, daemon=True
    )
    profiler_thread.start()


def _force_flush():
    profiler = get_profiler()
    profiler.force_flush()


def _start_profiling(opts):
    profiler = get_profiler()

    if profiler.is_running():
        logger.warning("profiler already running")
        return

    profiler.setup(
        opts.resource,
        opts.endpoint,
        opts.call_stack_interval_millis,
        opts.include_internal_stacks,
    )

    logger.debug(
        "starting profiling call_stack_interval_millis=%s endpoint=%s",
        opts.call_stack_interval_millis,
        opts.endpoint,
    )
    wrapt.wrap_function_wrapper(
        opentelemetry.context, "attach", profiler.make_wrapped_context_attach()
    )
    wrapt.wrap_function_wrapper(
        opentelemetry.context, "detach", profiler.make_wrapped_context_detach()
    )

    # Windows does not have register_at_fork
    if hasattr(os, "register_at_fork"):
        os.register_at_fork(after_in_child=lambda: _start_profiler_thread(profiler))
    atexit.register(stop_profiling)

    _start_profiler_thread(profiler)


def start_profiling(
    service_name: Optional[str] = None,
    resource_attributes: Optional[Dict[str, Union[str, bool, int, float]]] = None,
    endpoint: Optional[str] = None,
    call_stack_interval_millis: Optional[int] = None,
):
    resource = _create_resource(service_name, resource_attributes)
    opts = _Options(resource, endpoint, call_stack_interval_millis)
    _start_profiling(opts)


def stop_profiling():
    profiler = get_profiler()
    if not profiler.is_running():
        return

    profiler.running = False
    with profiler.condition:
        # Wake up the profiler thread
        profiler.condition.notify_all()
        # Wait for the profiler thread to exit
        profiler.condition.wait()

    unwrap(opentelemetry.context, "attach")
    unwrap(opentelemetry.context, "detach")



# Copyright Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Any, Dict

from opentelemetry.instrumentation.distro import BaseDistro  # type: ignore
from pkg_resources import EntryPoint

from splunk_otel.metrics import _configure_metrics
from splunk_otel.options import _Options
from splunk_otel.profiling import _start_profiling
from splunk_otel.profiling.options import _Options as ProfilingOptions
from splunk_otel.tracing import _configure_tracing
from splunk_otel.util import _is_truthy

logger = logging.getLogger(__name__)


class _SplunkDistro(BaseDistro):
    def __init__(self):
        tracing_enabled = os.environ.get("OTEL_TRACE_ENABLED", True)
        profiling_enabled = os.environ.get("SPLUNK_PROFILER_ENABLED", False)
        self._tracing_enabled = _is_truthy(tracing_enabled)
        self._profiling_enabled = _is_truthy(profiling_enabled)
        if not self._tracing_enabled:
            logger.info(
                "tracing has been disabled with OTEL_TRACE_ENABLED=%s", tracing_enabled
            )

        metrics_enabled = os.environ.get("OTEL_METRICS_ENABLED", True)
        self._metrics_enabled = _is_truthy(metrics_enabled)
        if not self._metrics_enabled:
            logger.info(
                "metering has been disabled with OTEL_METRICS_ENABLED=%s", metrics_enabled
            )

    def _configure(self, **kwargs: Dict[str, Any]) -> None:
        options = _Options()

        if self._tracing_enabled:
            _configure_tracing(options)

        if self._profiling_enabled:
            _start_profiling(ProfilingOptions(options.resource))

        if self._metrics_enabled:
            _configure_metrics()

    def load_instrumentor(self, entry_point: EntryPoint, **kwargs):
        if self._tracing_enabled:
            super().load_instrumentor(entry_point, **kwargs)



# Copyright Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import os.path
from argparse import REMAINDER, ArgumentParser
from logging import getLogger

from opentelemetry.instrumentation.auto_instrumentation import run as otel_run

from splunk_otel.version import _format_version_info

logger = getLogger(__file__)

ap = ArgumentParser()
ap.add_argument(
    "--token",
    "-t",
    required=False,
    type=str,
    dest="token",
    help="Your Splunk Access Token (SPLUNK_ACCESS_TOKEN env var by default)",
)
ap.add_argument(
    "--service-name",
    "-s",
    required=False,
    type=str,
    dest="service_name",
    help="The service name that should be passed to a tracer provider.",
)
ap.add_argument(
    "--version",
    "-v",
    required=False,
    action="store_true",
    dest="version",
    help="Print version information",
)
ap.add_argument(
    "command",
    help="Arguments for your application.",
    nargs=REMAINDER,
)


def run() -> None:
    args = ap.parse_args()
    if args.version:
        print(_format_version_info())
        return

    if args.token:
        os.environ["SPLUNK_ACCESS_TOKEN"] = args.token

    if args.service_name:
        os.environ["OTEL_SERVICE_NAME"] = args.service_name

    if not args.command:
        ap.error(ap.format_help())

    try:
        otel_run()
    except TypeError:
        logger.error("failed to execute program: %s", " ".join(args.command))
        raise


def run_deprecated() -> None:
    logger.warning(
        (
            "splk-py-trace is deprecated and will be removed in a future version. "
            "Please use `splunk-py-trace` instead."
        )
    )
    run()
